import kebabCase from 'lodash/kebabCase';
import merge from 'lodash/merge';
import camelCase from 'lodash/camelCase';
import isObject from 'lodash/isObject';
import { h, defineComponent, inject } from 'vue';
import uniq from 'lodash/uniq';
import omit from 'lodash/omit';

const props = [
    // Common
    ["display", null],
    ["position", null],
    "radius",
    "shadow",
    // Sizes
    ["width", "w"],
    ["w", "w"],
    ["height", "h"],
    ["h", "h"],
    ["maxW", "mw"],
    ["maxWidth", "mw"],
    ["maxH", "mh"],
    ["maxHeight", "mh"],
    // Flex
    ["alignItems", "items"],
    ["items", "items"],
    ["justifyContent", "justify"],
    ["flexWrap", "flex-wrap"],
    ["flexDirection", "flex"],
    ["direction", "flex"],
    ["dir", "flex"],
    ["wrap", "flex-wrap"],
    ["placeItems", "place"],
    ["place", "place"],
    // Spacing
    "gap",
    "gapX",
    "gapY",
    ["margin", "m"],
    ["marginX", "mx"],
    ["marginY", "my"],
    ["padding", "p"],
    ["paddingX", "px"],
    ["paddingY", "py"],
    "m",
    "mt",
    "mr",
    "mb",
    "ml",
    "mx",
    "my",
    "p",
    "pt",
    "pr",
    "pb",
    "pl",
    "px",
    "py",
    // Grid
    // TODO
    // Colors
    ["color", "text"],
    ["bg", "bg"],
    ["background", "bg"],
    ["borderColor", "border"],
    // Typography
    ["font", "font"],
    ["fontFamily", "font"],
    ["fontSize", "text"],
    ["fontWeight", "weight"],
    ["fontStyle", "font-style"],
    ["weight", "weight"],
    ["textAlign", "text"],
    ["textTransform", null],
    ["letterSpacing", "track"],
    ["lineHeight", "leading"],
];

function resolvePropConfig(config) {
    let propName = config;
    let classPrefix = null;
    let variants = true;
    if (typeof config === "string") {
        classPrefix = config;
    }
    else if (Array.isArray(config)) {
        propName = config[0];
        classPrefix = config[1] ?? config[0];
        variants = config[2] ?? true;
    }
    if (typeof classPrefix === "string") {
        classPrefix = kebabCase(classPrefix);
    }
    return {
        propName: camelCase(propName),
        classPrefix,
        variants,
    };
}
const tokens = {
    variantSeparator: ":",
    modifierSeparator: "--",
};
function buildClassName(prop, value = "", variant = null) {
    const utilityVariantPrefix = variant && variant !== "base"
        ? kebabCase(variant) + tokens.variantSeparator
        : "";
    const utilityPrefix = prop.classPrefix
        ? prop.classPrefix + tokens.modifierSeparator
        : "";
    const utilityValue = String(value).replace(".", "-");
    return [utilityVariantPrefix, utilityPrefix, utilityValue]
        .filter(Boolean)
        .join("");
}
function expandVariantPropClasses(propConfig, value) {
    let classes = [];
    if (isObject(value)) {
        for (let variant in value) {
            value = value;
            // Don't output 'falsey' values except for numbers
            if ([null, false, undefined].includes(value[variant])) {
                continue;
            }
            classes.push(buildClassName(propConfig, value[variant], variant));
        }
        return classes;
    }
    if (typeof value === "string") {
        value = String(value).replace(/\s+/g, " ").split(" ");
    }
    if (typeof value == "number")
        value = String(value);
    if (!Array.isArray(value))
        value = [value];
    value.forEach((val) => {
        let [variantOrValue, ...value] = val.split(":");
        if (value.length) {
            classes.push(buildClassName(propConfig, value.join(":"), variantOrValue));
        }
        else {
            classes.push(buildClassName(propConfig, variantOrValue));
        }
    });
    return classes;
}

const propKeys = [];
const propsMap = {};
/*
 * Map prop config to parser compatible config.
 */
props.forEach((config) => {
    // Normalize config short-hands
    const resolvedConfig = resolvePropConfig(config);
    // Record prop names so they can be plucked out and processed from
    // the `useBoxProps` hook.
    propKeys.push(resolvedConfig.propName);
    propsMap[resolvedConfig.propName] = resolvedConfig;
});
/**
 * Take an attribute and value; return a classlist to be applied to
 * a component
 */
function parser(prop, value) {
    const propKey = camelCase(prop);
    const propConfig = propsMap[propKey];
    let classes = [];
    console.log({ propKey, propConfig });
    // Variant aware props
    if (propConfig.variants) {
        classes = [...expandVariantPropClasses(propConfig, value)];
    }
    else {
        // One-off props
        classes.push(buildClassName(propConfig, value));
    }
    return classes;
}

var classnames = {exports: {}};

/*!
  Copyright (c) 2018 Jed Watson.
  Licensed under the MIT License (MIT), see
  http://jedwatson.github.io/classnames
*/

(function (module) {
/* global define */
(function () {

  var hasOwn = {}.hasOwnProperty;

  function classNames() {
    var classes = [];

    for (var i = 0; i < arguments.length; i++) {
      var arg = arguments[i];
      if (!arg) continue;
      var argType = typeof arg;

      if (argType === 'string' || argType === 'number') {
        classes.push(arg);
      } else if (Array.isArray(arg)) {
        if (arg.length) {
          var inner = classNames.apply(null, arg);

          if (inner) {
            classes.push(inner);
          }
        }
      } else if (argType === 'object') {
        if (arg.toString === Object.prototype.toString) {
          for (var key in arg) {
            if (hasOwn.call(arg, key) && arg[key]) {
              classes.push(key);
            }
          }
        } else {
          classes.push(arg.toString());
        }
      }
    }

    return classes.join(' ');
  }

  if (module.exports) {
    classNames.default = classNames;
    module.exports = classNames;
  } else {
    window.classNames = classNames;
  }
})();
}(classnames));

var classNames = classnames.exports;

const defaultConfig = {
    prefix: "r",
    boxProps: {
        parser,
        list: propKeys,
    },
    classNamer: function bem(baseClass) {
        return {
            block: () => baseClass,
            element: (element) => `${baseClass}__${element}`,
            modifier: (variants) => {
                return classNames(variants)
                    .split(" ")
                    .map((variant) => `${baseClass}--${variant}`)
                    .join(" ");
            },
        };
    },
};

function applyUtilities(attrs = {}, config) {
    let classes = [];
    // Flatten attributes into a class list compatible with vue's
    // v-bind:class binding
    for (let attr in attrs) {
        if (config.boxProps.list.indexOf(camelCase(attr)) !== -1 ||
            config.boxProps.list.indexOf(kebabCase(attr)) !== -1) {
            classes = [...classes, ...config.boxProps.parser(attr, attrs[attr])];
        }
    }
    // Remove polymorphic tag prop from attributes as these will be forwarded
    // directly to the resulting html tag
    delete attrs.as;
    delete attrs.is;
    delete attrs.tag;
    return [omit(attrs, config.boxProps.list), classes];
}

function render(defaultTag = "div", { slots, props, attrs, config, slot = "default" }) {
    let children = slots.default?.(slot);
    let tag = attrs.tag ?? attrs.as ?? attrs.is ?? defaultTag;
    let [attributes, utilityClasses] = applyUtilities(merge({}, attrs, props, { class: [attrs.class, props.class] }), config);
    // Compile classList and remove duplicate classes (for instance if a prop and a utility class were defined)
    const classes = uniq(classNames([props.class, attrs.class, utilityClasses]).split(" "));
    if (classes.length) {
        attributes.class = classes;
    }
    return h(tag, attributes, children);
}

const Box = defineComponent({
    inheritAttrs: false,
    setup(_, { slots, attrs }) {
        const config = inject("$riverConfig", defaultConfig);
        const bem = config.classNamer("box");
        return () => render("div", {
            config,
            slots,
            attrs,
            props: {
                class: [bem.block()],
            },
        });
    },
});

const Container = defineComponent({
    inheritAttrs: false,
    props: {
        size: [String, Object],
    },
    setup(props, { slots, attrs }) {
        const config = inject("$riverConfig", defaultConfig);
        const bem = config.classNamer("container");
        return () => render("div", {
            config,
            slots,
            attrs,
            props: {
                class: [bem.block()],
                maxWidth: `container-${props.size}`,
            },
        });
    },
});

const Heading = defineComponent({
    inheritAttrs: false,
    props: {
        size: {
            type: String,
            default: "1",
        },
    },
    setup(props, { slots, attrs }) {
        const config = inject("$riverConfig", defaultConfig);
        const bem = config.classNamer("heading");
        return () => render("h2", {
            config,
            slots,
            attrs,
            props: {
                class: [bem.block(), bem.modifier([props.size])],
            },
        });
    },
});

const Paragraph = defineComponent({
    inheritAttrs: false,
    setup(_, { slots, attrs }) {
        const config = inject("$riverConfig", defaultConfig);
        const bem = config.classNamer("paragraph");
        return () => render("p", {
            config,
            slots,
            attrs,
            props: {
                class: [bem.block()],
            },
        });
    },
});

const Row = defineComponent({
    inheritAttrs: false,
    setup(_, { slots, attrs }) {
        const config = inject("$riverConfig", defaultConfig);
        const bem = config.classNamer("row");
        return () => render("div", {
            config,
            slots,
            attrs,
            props: {
                class: [bem.block()],
            },
        });
    },
});

const Stack = defineComponent({
    inheritAttrs: false,
    setup(_, { slots, attrs }) {
        const config = inject("$riverConfig", defaultConfig);
        const bem = config.classNamer("stack");
        return () => render("div", {
            config,
            slots,
            attrs,
            props: {
                class: [bem.block()],
            },
        });
    },
});

const Text = defineComponent({
    inheritAttrs: false,
    setup(_, { slots, attrs }) {
        const config = inject("$riverConfig", defaultConfig);
        const bem = config.classNamer("text");
        return () => render("div", {
            config,
            slots,
            attrs,
            props: {
                class: [bem.block()],
            },
        });
    },
});

var components = /*#__PURE__*/Object.freeze({
  __proto__: null,
  Box: Box,
  Container: Container,
  Heading: Heading,
  Paragraph: Paragraph,
  Row: Row,
  Stack: Stack,
  Text: Text
});

function getComponentName(name, config) {
    return [config.prefix, kebabCase(name)].join("-").toLowerCase();
}
var plugin = {
    install(app, options = {}) {
        const config = merge({}, defaultConfig, options);
        app.config.globalProperties.$river = {
            config,
        };
        app.provide("$riverConfig", config);
        for (const component in components) {
            // @ts-ignore-next
            app.component(getComponentName(component, config), components[component]);
        }
    },
};

export { applyUtilities, plugin as default, defaultConfig, props as defaultProps, parser, plugin, propKeys, propsMap };
