import kebabCase from "lodash/kebabCase";
import camelCase from "lodash/camelCase";
import isObject from "lodash/isObject";
import { PropConfig, BoxPropConfig } from "../types";

export function resolvePropConfig(config: PropConfig): BoxPropConfig {
  let propName = config;
  let classPrefix: string | null = null;
  let variants = true;

  if (typeof config === "string") {
    classPrefix = config;
  } else if (Array.isArray(config)) {
    propName = config[0];
    classPrefix = config[1] ?? config[0];
    variants = config[2] ?? true;
  }

  if (typeof classPrefix === "string") {
    classPrefix = kebabCase(classPrefix);
  }

  return {
    propName: camelCase(propName as string),
    classPrefix,
    variants,
  };
}

const tokens = {
  variantSeparator: ":",
  modifierSeparator: "--",
};

export function buildClassName(
  prop: BoxPropConfig,
  value = "",
  variant: string | null = null
) {
  const utilityVariantPrefix =
    variant && variant !== "base"
      ? kebabCase(variant) + tokens.variantSeparator
      : "";

  const utilityPrefix = prop.classPrefix
    ? prop.classPrefix + tokens.modifierSeparator
    : "";

  const utilityValue = String(value).replace(".", "-");

  return [utilityVariantPrefix, utilityPrefix, utilityValue]
    .filter(Boolean)
    .join("");
}

export function expandVariantPropClasses(
  propConfig: BoxPropConfig,
  value: any
) {
  let classes: any[] = [];

  if (isObject(value)) {
    for (let variant in value) {
      value = value as Record<string, any>;

      // Don't output 'falsey' values except for numbers
      if ([null, false, undefined].includes(value[variant])) {
        continue;
      }

      classes.push(buildClassName(propConfig, value[variant], variant));
    }
    return classes;
  }

  if (typeof value === "string") {
    value = String(value).replace(/\s+/g, " ").split(" ");
  }
  if (typeof value == "number") value = String(value);
  if (!Array.isArray(value)) value = [value];

  value.forEach((val: string) => {
    let [variantOrValue, ...value] = val.split(":");

    if (value.length) {
      classes.push(buildClassName(propConfig, value.join(":"), variantOrValue));
    } else {
      classes.push(buildClassName(propConfig, variantOrValue));
    }
  });

  return classes;
}

export function normalizeClassAsArray(classNames?: any): any[] {
  if (!classNames) {
    return [];
  }
  if (typeof classNames === "object" && !Array.isArray(classNames)) {
    return [classNames];
  }
  if (typeof classNames === "string") {
    return [classNames];
  }
  return classNames;
}
